import bpy

from .. utils import (
	addon,
	theme_utils,
)

def linked_color_prop(layout: bpy.types.UILayout, data: bpy.types.PointerProperty, property: str ) -> None:
	'''
	Adds a small toggle to the right of the color property to link it 
	to the corresponding color property (determined elsewhere).
	
	'''
	
	row = layout.row(align=True)
	row.prop(data, property)

	row.prop(
		data, 
		property+'_link', 
		text='', 
		icon='LINKED' if eval(f'data.{property}_link') else 'UNLINKED'
	)


# MAIN PANEL:
# ===============================

class THEMECANDY_PT_MainPanel(bpy.types.Panel):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_category = 'Theme Candy'
	bl_label = 'Theme Candy'
	# bl_options = {'HIDE_HEADER'}

	def draw(self, context):
		layout = self.layout
		col = layout.column(align=True)

		# RELOAD COLORS:
		# ===============================
		col.operator('theme_candy.load_theme_colors', text='Refresh Panel', icon='FILE_REFRESH')
		col.separator()

		# PRESET DROPDOWN:
		# ===============================
		row = col.row(align=True)
		row.menu('USERPREF_MT_theme_candy_presets', text=theme_utils.get_theme_name())
		row.operator('wm.interface_theme_preset_add',	 text='', icon='ADD')
		row.operator('wm.interface_theme_preset_remove', text='', icon='REMOVE')
		row.operator('wm.interface_theme_preset_save',	 text='', icon='FILE_TICK')
		col.separator()

		col.operator('theme_candy.reset_default_theme', text='Reset to Default Theme')
		col.separator()

		# CUSTOMIZATION LEVEL:
		# ===============================
		col.label(text='Customization Level:')
		row = col.row(align=True)
		row.prop(addon.prefs(), 'mode', expand=True)
		col.separator()

		# SUB PANELS GET ADDED HERE BY BLENDER:
		# ===============================
		# ...


# SUB PANELS:
# ===============================

class THEMECANDY_PT_UserInterface(bpy.types.Panel):
	bl_label = 'User Interface'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_MainPanel'
	# bl_options = {'DEFAULT_CLOSED'} if addon.prefs().collapse_panels else {}	# Does not work because addon prfs do not exist when this class registers.

	def draw(self, context):
		candy_props = context.window_manager.themecandy
		advanced_mode = addon.prefs().mode == 'ADVANCED'

		layout = self.layout
		layout.use_property_split = True

		col = layout.column(align=True)
		col.prop(candy_props, 'ui_selected')
		col.prop(candy_props, 'text_selected')
		col.separator()
		if advanced_mode:
			col.prop(candy_props, 'ui_unselected')
			col.prop(candy_props, 'text_unselected')
			col.separator()
			col.prop(candy_props, 'text_shadow_size')
			col.prop(candy_props, 'text_shadow_alpha')
			col.separator()

			linked_color_prop(layout=col, data=candy_props, property='modifier_active')

		col.prop(candy_props, 'roundness')


class THEMECANDY_PT_ViewportSelections(bpy.types.Panel):
	bl_label = 'Viewport Selections'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_MainPanel'

	def draw(self, context):
		candy_props = context.window_manager.themecandy
		advanced_mode = addon.prefs().mode == 'ADVANCED'

		layout = self.layout
		layout.use_property_split = True

		col = layout.column(align=True)

		if advanced_mode:
			linked_color_prop(layout=col, data=candy_props, property='object_active')
			col.prop(candy_props, 'object_select')
			col.separator()

			row = col.row(align=True)
			row.prop(candy_props, 'vertex')
			linked_color_prop(layout=col, data=candy_props, property='vertex_select')
			linked_color_prop(layout=col, data=candy_props, property='edge_select')
			linked_color_prop(layout=col, data=candy_props, property='edge_mode_select')
			linked_color_prop(layout=col, data=candy_props, property='face_select')
			linked_color_prop(layout=col, data=candy_props, property='face_mode_select')
			col.separator()

			linked_color_prop(layout=col, data=candy_props, property='element_active')
			col.separator()

			col.prop(candy_props, 'vertex_size')
		else:
			linked_color_prop(layout=col, data=candy_props, property='global_select')
			row = col.row(align=True)
			row.prop(candy_props, 'face_alpha')
		

class THEMECANDY_PT_ViewportBackground(bpy.types.Panel):
	bl_label = 'Viewport Background'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_MainPanel'

	def draw(self, context):
		candy_props = context.window_manager.themecandy
		advanced_mode = addon.prefs().mode == 'ADVANCED'

		layout = self.layout
		layout.use_property_split = True

		col = layout.column(align=True)

		col.prop(candy_props, 'background_type')
		col.prop(candy_props, 'gradient_high', text='Color' if candy_props.background_type == 'SINGLE_COLOR' else 'Gradient High')
		if candy_props.background_type != 'SINGLE_COLOR':
			col.prop(candy_props, 'gradient_low')
		
		col.separator()

		col.prop(candy_props, 'grid_color')
		col.separator()

		if advanced_mode:
			linked_color_prop(layout=col, data=candy_props, property='axis_x')
			col.prop(candy_props, 'axis_y')
			col.prop(candy_props, 'axis_z')


class THEMECANDY_PT_Panels(bpy.types.Panel):
	bl_label = 'Panels'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_UserInterface'
	bl_options = {'DEFAULT_CLOSED'}

	@classmethod
	def poll(self, context):
		return addon.prefs().mode == 'ADVANCED'

	def draw(self, context):
		candy_props = context.window_manager.themecandy
		advanced_mode = addon.prefs().mode == 'ADVANCED'

		layout = self.layout
		layout.use_property_split = True

		col = layout.column(align=True)
		
		row = col.row(align=True)
		row.prop(candy_props, 'outliner')
		# row.separator()
		# row.prop(candy_props, 'outliner_panel', text='')	# Outliner has no Panels.

		row = col.row(align=True)
		row.prop(candy_props, 'properties')
		row.separator()
		row.prop(candy_props, 'properties_panel', text='')

		row = col.row(align=True)
		row.prop(candy_props, 'preferences')
		row.separator()
		row.prop(candy_props, 'preferences_panel', text='')

		row = col.row(align=True)
		row.prop(candy_props, 'node_editor_grid')
		row.separator()
		row.prop(candy_props, 'node_editor', text='')
		row.separator()
		row.prop(candy_props, 'node_editor_panel', text='')
		
		row = col.row(align=True)
		row.prop(candy_props, 'image_editor')
		row.separator()
		row.prop(candy_props, 'image_editor_panel', text='')

		
		row = col.row(align=True)
		row.prop(candy_props, 'console_header')
		row.separator()
		row.prop(candy_props, 'console', text='')
		row.separator()
		row.prop(candy_props, 'console_panel', text='')
		

		row = col.row(align=True)
		row.prop(candy_props, 'text_editor_header')
		row.separator()
		row.prop(candy_props, 'text_editor', text='')
		row.separator()
		row.prop(candy_props, 'text_editor_panel', text='')

		row = col.row(align=True)
		row.prop(candy_props, 'dopesheet_editor')
		# row.separator()
		# row.prop(candy_props, 'dopesheet_editor_panel', text='')

		# col.prop(candy_props, 'sequence_editor')
		# col.prop(candy_props, 'spreadsheet')
		# col.prop(candy_props, 'info', text=)
		# col.prop(candy_props, 'clip_editor', text=)
		# col.prop(candy_props, 'nla_editor', text=)
		# col.prop(candy_props, 'file_browser', text=)
		# col.prop(candy_props, 'graph_editor', text=)

		col.separator()

		col.prop(candy_props, 'editor_outline')


class THEMECANDY_PT_Outliner(bpy.types.Panel):
	bl_label = 'Outliner'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_MainPanel'

	def draw(self, context):
		candy_props = context.window_manager.themecandy
		advanced_mode = addon.prefs().mode == 'ADVANCED'

		layout = self.layout
		layout.use_property_split = True

		col = layout.column(align=True)

		linked_color_prop(layout=col, data=candy_props, property='outliner_active_highlight')	
		linked_color_prop(layout=col, data=candy_props, property='outliner_active_text')

		if advanced_mode:
			col.separator()
			col.prop(candy_props, 'outliner_selected_highlight')
			col.prop(candy_props, 'outliner_selected_text')

		col.separator()

		col.prop(candy_props, 'outliner_zebra')


class THEMECANDY_PT_Icons(bpy.types.Panel):
	bl_label = 'Icons'
	# bl_idname = 'OBJECT_PT_sub_hello'
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_parent_id = 'THEMECANDY_PT_UserInterface'
	# bl_context = 'scene'
	bl_options = {'DEFAULT_CLOSED'}

	@classmethod
	def poll(self, context):
		return addon.prefs().mode == 'ADVANCED'

	def draw(self, context):
		candy_props = context.window_manager.themecandy

		layout = self.layout
		layout.use_property_split = True
		col = layout.column(align=True)

		col.prop(candy_props, 'icon_scene')
		col.prop(candy_props, 'icon_collection')
		col.prop(candy_props, 'icon_object')
		col.prop(candy_props, 'icon_object_data')
		col.prop(candy_props, 'icon_modifier')
		col.prop(candy_props, 'icon_shading')
		col.prop(candy_props, 'icon_folder')



classes = (
	THEMECANDY_PT_MainPanel,
	THEMECANDY_PT_UserInterface,
	THEMECANDY_PT_ViewportSelections,
	THEMECANDY_PT_ViewportBackground,
	THEMECANDY_PT_Panels,
	THEMECANDY_PT_Outliner,
	THEMECANDY_PT_Icons,
)

def toggle_n_panel(self, context):
        for cls in classes:
            if self.show_n_panel:
                bpy.utils.register_class(cls)
            else:
                bpy.utils.unregister_class(cls)


def register():
	for cls in classes:
		bpy.utils.register_class(cls)


def unregister():
	for cls in classes:
		bpy.utils.unregister_class(cls)