import bpy
from mathutils import Color, Vector

from .. utils import (
	addon,
)


class SharedExecute:
	def execute(self, context):
		if addon.prefs().lock_updates:
			return
		
		addon.prefs().lock_updates = True

		override_linked_colors(context)
		set_ui(context)
		set_view3d(context)
		set_outliner(context)

		addon.prefs().lock_updates = False


def get_theme_name() -> str:
	'''
	Returns the label of the loaded theme.
	'''
	
	preset_menu_class = getattr(bpy.types, 'USERPREF_MT_interface_theme_presets', None)
	
	if preset_menu_class is None:
		return 'Presets'
	
	return getattr(preset_menu_class, 'bl_label', 'Presets')


def get_theme_vars(context):
	'''
	Returns: ui, view3d, properties, outliner, text.
	'''

	theme = context.preferences.themes[0]
	
	ui          = theme.user_interface
	view3d      = theme.view_3d
	ui_props    = theme.properties
	outliner    = theme.outliner
	text        = context.preferences.ui_styles[0].widget

	return ui, view3d, ui_props, outliner, text

def get_ui_elements() -> dict:
	'''
	- Everything listed here will be affected by the <roundness> property.
	- Only elements tagged with 'colored' can be affected by color changes.
	'''

	return {
		'wcol_box'          : '',
		'wcol_list_item'    : '',
		'wcol_menu'         : 'colored',
		'wcol_menu_back'    : '',           # Do Not color this one, do it manually in the special cases if necessary.
		'wcol_menu_item'    : 'colored',
		'wcol_num'          : '',
		'wcol_numslider'    : '',           # special colored case, leave empty.
		'wcol_option'       : '',           # special colored case, leave empty.
		'wcol_pie_menu'     : 'colored',
		'wcol_progress'     : 'colored',
		'wcol_pulldown'     : 'colored',
		'wcol_radio'        : 'colored',
		'wcol_regular'      : 'colored',
		'wcol_scroll'       : '',
		# 'wcol_state'      : '',           # really different from other <wcol> candy_props, comment for now.
		'wcol_tab'          : '',
		'wcol_text'         : '',
		'wcol_toggle'       : 'colored',
		'wcol_tool'         : 'colored',
		'wcol_toolbar_item' : 'colored',
		'wcol_tooltip'      : 'colored',    # Meaningless for now but we.
	}

def get_unselected_ui_elements() -> dict:

	return {
		# 'wcol_box'          : '',
		# 'wcol_list_item'    : '',
		# 'wcol_menu'         : 'colored',
		# 'wcol_menu_back'    : '',           # Do Not color this one, do it manually in the special cases if necessary.
		'wcol_menu_item'    : 'colored',
		'wcol_num'          : '',
		'wcol_numslider'    : '',           # special colored case, leave empty.
		# 'wcol_option'       : '',           # special colored case, leave empty.
		# 'wcol_pie_menu'     : 'colored',
		'wcol_progress'     : 'colored',
		'wcol_pulldown'     : 'colored',
		'wcol_radio'        : 'colored',
		'wcol_regular'      : 'colored',
		# 'wcol_scroll'       : '',
		# 'wcol_state'      : '',           # really different from other <wcol> candy_props, comment for now.
		# 'wcol_tab'          : '',
		# 'wcol_text'         : '',
		'wcol_toggle'       : 'colored',
		'wcol_tool'         : 'colored',
		# 'wcol_toolbar_item' : 'colored',
		# 'wcol_tooltip'      : 'colored',    # Maybe we add a tooltip and menu background colored property later.
	}

def get_theme_spaces() -> list:
	'''
	Just a list I.keep for reference. Not used in the code.
	'''

	return [
		# 'user_interface',
		'dopesheet_editor',
		'file_browser',
		'graph_editor',
		'image_editor',
		'info',
		'clip_editor',
		'node_editor',
		'nla_editor',
		'outliner',
		'preferences',
		'properties',
		'console',
		'spreadsheet',
		# 'statusbar',
		'text_editor',
		# 'topbar',
		'sequence_editor',

	]


def override_linked_colors(context):
	'''
	Some colors can be linked to the UI color with a toggle. 
	Always run BEFORE modifying the current theme.
	'''

	candy_props = context.window_manager.themecandy

	if candy_props.object_active_link:    candy_props.object_active    = candy_props.ui_selected[:3]
	if candy_props.vertex_select_link:    candy_props.vertex_select    = candy_props.ui_selected[:3]
	if candy_props.edge_select_link:      candy_props.edge_select      = candy_props.ui_selected[:3]
	if candy_props.edge_mode_select_link: candy_props.edge_mode_select = candy_props.ui_selected[:3]
	if candy_props.face_select_link:      candy_props.face_select      = candy_props.ui_selected
	if candy_props.face_mode_select_link: candy_props.face_mode_select = candy_props.ui_selected
	if candy_props.element_active_link:   candy_props.element_active   = candy_props.ui_selected
	if candy_props.global_select_link:    candy_props.global_select    = candy_props.ui_selected[:3]
	if candy_props.axis_x_link:           candy_props.axis_x           = candy_props.ui_selected[:3]
	if candy_props.modifier_active_link:  candy_props.modifier_active  = candy_props.ui_selected

	if candy_props.outliner_active_highlight_link: candy_props.outliner_active_highlight = candy_props.ui_selected[:3]
	if candy_props.outliner_active_text_link:      candy_props.outliner_active_text      = candy_props.text_selected[:3]



def init_color_properties(context):
	'''
	Sets the operator's color properties to match the current theme colors
	(usefeful when the operator's main panel is drawn for the first time).

	DO NOT ALTER THE CURRENT THEME FROM HERE!!!
	'''

	addon.prefs().lock_updates = True

	ui, view3d, uiprops, outliner, text = get_theme_vars(context)
	
	candy_props = context.window_manager.themecandy

	# MAKE SURE THESE ARE SET FIRST BECAUSE OTHER VALUES CAN BE LINKED TO THEM
	candy_props.ui_selected      = ui.wcol_regular.inner_sel
	candy_props.text_selected    = ui.wcol_regular.text_sel
	candy_props.ui_unselected    = ui.wcol_regular.inner
	candy_props.text_unselected  = ui.wcol_regular.text
	candy_props.roundness        = ui.wcol_regular.roundness

	candy_props.global_select    = view3d.object_active
	candy_props.object_active    = view3d.object_active
	candy_props.object_select    = view3d.object_selected
	candy_props.vertex           = view3d.vertex
	candy_props.vertex_select    = view3d.vertex_select
	candy_props.edge_select      = view3d.edge_select
	candy_props.edge_mode_select = view3d.edge_mode_select	# BL 4.1
	candy_props.face_select      = view3d.face_select
	candy_props.face_mode_select = view3d.face_mode_select	# BL 4.1
	candy_props.face_alpha       = view3d.face_mode_select[3]	# BL 4.1

	candy_props.element_active   = view3d.editmesh_active

	candy_props.background_type  = view3d.space.gradients.background_type
	candy_props.gradient_high    = view3d.space.gradients.high_gradient
	candy_props.gradient_low     = view3d.space.gradients.gradient
	candy_props.grid_color       = view3d.grid
	candy_props.vertex_size      = view3d.vertex_size

	candy_props.axis_x = ui.axis_x
	candy_props.axis_y = ui.axis_y
	candy_props.axis_z = ui.axis_z

	candy_props.text_shadow_size  = text.shadow
	candy_props.text_shadow_alpha = text.shadow_alpha
	
	candy_props.outliner_active_highlight   = outliner.active
	candy_props.outliner_active_text        = outliner.active_object
	candy_props.outliner_selected_highlight = outliner.selected_highlight
	candy_props.outliner_selected_text      = outliner.selected_object
	candy_props.outliner_zebra              = outliner.row_alternate[3] != 0

	candy_props.modifier_active = uiprops.active_modifier
	if addon.prefs().mode == 'BASIC':
		candy_props.modifier_active = candy_props.ui_selected
	

	# Icons
	# ===================================================
	candy_props.icon_scene 		= ui.icon_scene
	candy_props.icon_collection 	= ui.icon_collection
	candy_props.icon_object 	= ui.icon_object
	candy_props.icon_object_data 	= ui.icon_object_data
	candy_props.icon_modifier 	= ui.icon_modifier
	candy_props.icon_shading 	= ui.icon_shading
	candy_props.icon_folder		= ui.icon_folder


	# Areas
	# ===================================================
	theme = context.preferences.themes[0]
	
	candy_props.node_editor			= theme.node_editor.space.back
	candy_props.node_editor_panel		= theme.node_editor.space.panelcolors.back
	candy_props.node_editor_grid		= theme.node_editor.grid

	candy_props.image_editor		= theme.image_editor.space.back
	candy_props.image_editor_panel		= theme.image_editor.space.panelcolors.back

	candy_props.outliner			= theme.outliner.space.back
	candy_props.outliner_panel		= theme.outliner.space.panelcolors.back

	candy_props.preferences			= theme.preferences.space.back
	candy_props.preferences_panel		= theme.preferences.space.panelcolors.back

	candy_props.properties			= theme.properties.space.back
	candy_props.properties_panel		= theme.properties.space.panelcolors.back
	
	candy_props.text_editor			= theme.text_editor.space.back
	candy_props.text_editor_header		= theme.text_editor.space.header
	candy_props.text_editor_panel		= theme.text_editor.space.panelcolors.back

	candy_props.console			= theme.console.space.back
	candy_props.console_header		= theme.console.space.header
	candy_props.console_panel		= theme.console.space.panelcolors.back

	candy_props.dopesheet_editor		= theme.dopesheet_editor.space.back
	# candy_props.dopesheet_editor_panel	= theme.dopesheet_editor.space.button

	candy_props.editor_outline 		= ui.editor_outline

	# Overrides for `Basic Mode` where user controls multiple element colors with a single color property.
	# ===================================================
	override_linked_colors(context)
	addon.prefs().lock_updates = False
	
	
def set_ui(context):
	candy_props = context.window_manager.themecandy
	ui, _, uiprops, _, text = get_theme_vars(context)

	if addon.prefs().mode == 'BASIC':
		candy_props.modifier_active = candy_props.ui_selected

	# UI
	# ===================================================
	for e in get_ui_elements():
		exec(f'ui.{e}.roundness = candy_props.roundness')

	colored_ui_elements = [key for key, val in get_ui_elements().items() if val == 'colored']
	for e in colored_ui_elements:
		exec(f'ui.{e}.inner_sel = candy_props.ui_selected')
		exec(f'ui.{e}.text_sel = candy_props.text_selected')

	
	for e in get_unselected_ui_elements().keys():
		exec(f'ui.{e}.inner = candy_props.ui_unselected')
		exec(f'ui.{e}.text = candy_props.text_unselected')

	# SPECIAL CASES
	# (Some elements require special colors for different reasons):
	# ===================================================

	if bpy.app.version >= (4, 0, 0):
		# Weird new way to color hovered menu items (EX: SHIFT+A). KEEP ALPHA at 0.0
		# It gets combined with text color now so you can never get the exact color you want LOL
		# And it affect other stuff like category labels in the search menu (EX : 'Mesh" when searching for a cube primitive.)
		# Fuck you Spiderman!
		ui.wcol_menu_item.inner = (*candy_props.ui_selected[:3], 0.0)	

	ui.wcol_text.item     = *candy_props.outliner_active_highlight * 1.47, 1   # Selected text highlignt.
	ui.wcol_num.item      = *candy_props.outliner_active_highlight * 1.47, 1   # Selected number highlignt.
	ui.widget_text_cursor =  candy_props.outliner_active_highlight * 2.2

	ui.wcol_progress.item  = candy_props.ui_selected   # Don't know what this is, but it has color.
	ui.wcol_list_item.item = candy_props.ui_selected   # Don't know what this is, but it has color.

	# Boolean Checkboxes.
	ui.wcol_option.inner_sel =  candy_props.ui_selected
	ui.wcol_option.item      = *candy_props.text_selected, 1

	ui.wcol_numslider.item     = candy_props.ui_selected
	ui.wcol_numslider.text_sel = candy_props.text_selected
	if addon.prefs().neutral_sliders:
		ui.wcol_numslider.item = (.2, .2, .2, 1)
		ui.wcol_numslider.text_sel = (1, 1, 1)


	# Text:
	# ===================================================
	text.shadow       = candy_props.text_shadow_size
	text.shadow_alpha = candy_props.text_shadow_alpha


	# Properties
	# ===================================================
	uiprops.active_modifier = candy_props.modifier_active


	# Axis and Gizmos:
	# ===================================================
	ui.axis_x = candy_props.axis_x
	ui.axis_y = candy_props.axis_y
	ui.axis_z = candy_props.axis_z


	# Icons:
	# ===================================================
	ui.icon_scene 		= candy_props.icon_scene 
	ui.icon_collection 	= candy_props.icon_collection
	ui.icon_object 		= candy_props.icon_object
	ui.icon_object_data 	= candy_props.icon_object_data
	ui.icon_modifier 	= candy_props.icon_modifier
	ui.icon_shading 	= candy_props.icon_shading
	ui.icon_folder 		= candy_props.icon_folder


	# Areas
	# ===================================================
	theme = context.preferences.themes[0]

	theme.node_editor.grid 				= candy_props.node_editor_grid
	theme.node_editor.space.back			=  candy_props.node_editor
	theme.node_editor.space.header			= *candy_props.node_editor, 1
	theme.node_editor.space.panelcolors.header	=  candy_props.node_editor_panel
	theme.node_editor.space.panelcolors.back	=  candy_props.node_editor_panel

	theme.image_editor.space.back			=  candy_props.image_editor
	theme.image_editor.space.header			= *candy_props.image_editor, 1
	theme.image_editor.space.panelcolors.header	=  candy_props.image_editor_panel
	theme.image_editor.space.panelcolors.back	=  candy_props.image_editor_panel
	
	theme.outliner.space.back			=  candy_props.outliner
	theme.outliner.space.header			= *candy_props.outliner, 1
	theme.outliner.space.panelcolors.header		=  candy_props.outliner_panel
	theme.outliner.space.panelcolors.back		=  candy_props.outliner_panel

	theme.preferences.space.back			=  candy_props.preferences
	theme.preferences.space.header			= *candy_props.preferences, 1
	theme.preferences.space.navigation_bar		= *candy_props.preferences, 1
	theme.preferences.space.execution_buts		= *candy_props.preferences, 1
	theme.preferences.space.panelcolors.header	=  candy_props.preferences_panel
	theme.preferences.space.panelcolors.back	=  candy_props.preferences_panel

	theme.properties.space.back			=  candy_props.properties
	theme.properties.space.header			= *candy_props.properties, 1
	theme.properties.space.panelcolors.header	=  candy_props.properties_panel
	theme.properties.space.panelcolors.back		=  candy_props.properties_panel
	# OFFSET BASED
	# theme.properties.space.panelcolors.header	= Vector((*candy_props.properties, 1)) + Vector((0.051, 0.051, 0.051, 0.051))
	# theme.properties.space.panelcolors.back		= Vector((*candy_props.properties, 1)) + Vector((0.051, 0.051, 0.051, 0.051))

	theme.text_editor.space.back			=  candy_props.text_editor
	theme.text_editor.space.header			=  candy_props.text_editor_header
	theme.text_editor.space.button			=  candy_props.text_editor_header	# The property name is 'button' but the label is 'Region Background'.
	theme.text_editor.space.panelcolors.header	=  candy_props.text_editor_panel
	theme.text_editor.space.panelcolors.back	=  candy_props.text_editor_panel

	theme.console.space.back			=  candy_props.console
	theme.console.space.header			=  candy_props.console_header
	theme.console.space.panelcolors.header		=  candy_props.console_panel
	theme.console.space.panelcolors.back		=  candy_props.console_panel

	theme.dopesheet_editor.space.back		=  candy_props.dopesheet_editor
	theme.dopesheet_editor.space.header		= *candy_props.dopesheet_editor, 1
	# theme.dopesheet_editor.space.button		=  candy_props.dopesheet_editor_panel
	theme.dopesheet_editor.space.panelcolors.header	=  candy_props.dopesheet_editor_panel
	theme.dopesheet_editor.space.panelcolors.back	=  candy_props.dopesheet_editor_panel

	ui.editor_outline = candy_props.editor_outline


def set_view3d(context):
	candy_props = context.window_manager.themecandy
	view3d = context.preferences.themes[0].view_3d

	if addon.prefs().mode == 'BASIC':

		# MIGHT WANT TO ADD THE NEW PROPERTIES FOR EDGE AND FACE MODE SELECT!!!
		candy_props.object_active    =  candy_props.global_select
		candy_props.object_select    =  candy_props.global_select
		candy_props.vertex_select    =  candy_props.global_select
		candy_props.edge_select      =  candy_props.global_select
		candy_props.edge_mode_select =  candy_props.global_select
		candy_props.face_select      = *candy_props.global_select, 0.2
		candy_props.face_mode_select = *candy_props.global_select, 0.2

		offset_view3d_colors(context)

	# Viewport Unselected:
	# ===================================================
	view3d.vertex = candy_props.vertex

	# Viewport Selections:
	# ===================================================
	view3d.object_active    = candy_props.object_active
	view3d.object_selected  = candy_props.object_select
	view3d.vertex_select    = candy_props.vertex_select
	view3d.edge_select      = candy_props.edge_select
	view3d.edge_mode_select = candy_props.edge_mode_select	# BL 4.1
	view3d.face_select      = candy_props.face_select
	view3d.face_mode_select = candy_props.face_mode_select	# BL 4.1
	view3d.editmesh_active  = candy_props.element_active

	# Viewport Background:
	# ===================================================

	view3d.space.gradients.background_type = candy_props.background_type
	view3d.space.gradients.high_gradient   = candy_props.gradient_high
	view3d.space.gradients.gradient        = candy_props.gradient_low

	view3d.grid = candy_props.grid_color

	# Other:
	# ===================================================

	view3d.vertex_size = candy_props.vertex_size


def calculate_offset_colors():
	'''
	Copy and Paste this function into Blender's Text Editor and it will print the color offsets
	between the Active Object color and the other View3D selection colors.
	'''
	
	import bpy
	from mathutils import Vector

	context = bpy.context
	theme = context.preferences.themes[0]
	view3d = theme.view_3d

	parent_color = view3d.object_active

	color_keys = [
		'object_selected',
		'vertex_select',
		'edge_select',
		'edge_mode_select',
		'face_select',
		'face_mode_select',
		# 'editmesh_active',
	]

	color_offsets = dict.fromkeys(color_keys)
	for key in color_offsets:
		color_offsets[key] = Vector(parent_color) - Vector(getattr(view3d, key)[:3])

	# These 'len' values are just for formatting the print output.
	longest_key = max(color_keys, key=len)
	max_len = len(longest_key)
	
	print()
	for key, val in color_offsets.items():
		print(key + ' '*(max_len-len(key)), val)


def offset_view3d_colors(context):
	'''
	Allows the user to control multiple 3D View colors with a single color proparty: <global_select>.
	Simulates the slight color variations in Blender's Default Dark Theme.
	'''

	candy_props = context.window_manager.themecandy

	# How much is each color offset from the Active Object color.
	# You get these values from the calculate_offset_colors() function.
	color_offsets = {
		'object_select'    : Color((0.0706,  0.2863, 0.1569)),
		'vertex_select'    : Color((0.0000,  0.1490, 0.1569)),
		'edge_select'      : Color((0.0000,  0.0275, 0.1569)),
		'edge_mode_select' : Color((0.0000, -0.2196, 0.1569)),
		'face_select'      : Color((0.0000, -0.0118, 0.1569)),
		'face_mode_select' : Color((0.0000, -0.0902, 0.1569)),
	}
	
	candy_props.object_select      -= color_offsets['object_select']
	candy_props.vertex_select      -= color_offsets['vertex_select' ]
	candy_props.edge_select        -= color_offsets['edge_select']
	candy_props.edge_mode_select   -= color_offsets['edge_mode_select']
	# Need to use Vectors because of the Alpha Value.
	candy_props.face_select         = Vector(candy_props.face_select)      - Vector((*color_offsets['face_select'],      0))
	candy_props.face_mode_select    = Vector(candy_props.face_mode_select) - Vector((*color_offsets['face_mode_select'], 0))
	
	candy_props.face_select[3]      = candy_props.face_alpha
	candy_props.face_mode_select[3] = candy_props.face_alpha

	# Special HSV cases for low Saturation colors with high/low Values.
	sat = candy_props.object_active.s
	val = candy_props.object_active.v
	
	for key in color_offsets:
		
		temp = getattr(candy_props, key)
		if sat <= 0.15:
			if isinstance(temp, Color):
				temp.s = 0.0
				setattr(candy_props, key, temp)
			else:
				r, g, b, a = temp
				temp = Color((r, g, b))
				temp.s = 0.0
				setattr(candy_props, key, (*temp, a))

			if key != 'object_select':
				continue

			temp = getattr(candy_props, key)
			if val >= 0.7:
				if isinstance(temp, Color):
					temp.v = 1 - candy_props.object_active.v
					setattr(candy_props, key, temp)
				else:
					r, g, b, a = temp
					temp = Color((r, g, b))
					temp.v = 1 - candy_props.object_active.v
					setattr(candy_props, key, (*temp, a))
				
			temp = getattr(candy_props, key)
			if val <= 0.3:
				if isinstance(temp, Color):
					temp.v = 1 - candy_props.object_active.v
					setattr(candy_props, key, temp)
				else:
					r, g, b, a = temp
					temp = Color((r, g, b))
					temp.v = 1 - candy_props.object_active.v
					setattr(candy_props, key, (*temp, a))


def offset_outliner_colors(context):
	'''
	Try to maintain the default Outliner color offsets between Active and Selected elements n Blender's Default Theme.
	It looks very stupid but it works as intended.
	'''

	candy_props = context.window_manager.themecandy

	# How much is each color offset from the Active Outliner elements.
	highlight_offset = Color((-.086, -.110, -.200))
	text_offset      = Color((-.071, -.313, -.157))

	# # Special offset overrides for non-saturated user colors.
	sat = candy_props.outliner_active_highlight.s
	if sat <= 0.3:
		highlight_offset = Color((-.14, -.14, -.14))

	if candy_props.outliner_active_text.s <= 0.2:
		text_offset = Color((-.17, -.17, -.17))

	candy_props.outliner_selected_highlight = candy_props.outliner_active_highlight + highlight_offset
	candy_props.outliner_selected_text      = candy_props.outliner_active_text      + text_offset


def set_outliner(context):
	candy_props = context.window_manager.themecandy
	outliner = context.preferences.themes[0].outliner

	if addon.prefs().mode == 'BASIC':
		# addon.prefs().lock_updates = True
		offset_outliner_colors(context)
		# addon.prefs().lock_updates = False

	outliner.active             = candy_props.outliner_active_highlight
	outliner.active_object      = candy_props.outliner_active_text
	outliner.selected_highlight = candy_props.outliner_selected_highlight 
	outliner.selected_object    = candy_props.outliner_selected_text

	outliner.row_alternate = (1, 1, 1, .016) if candy_props.outliner_zebra else (1, 1, 1, 0)

# def set_grease_pencil():
#	candy_props = context.window_manager.themecandy
# 	try:
# 		bpy.data.grease_pencils[0].layers[0].color = candy_props.grease_pencil
# 	except Exception:
# 		pass

